use crate::robot::robot_connector::{Error, RobotConnector};
use crate::robot::ROBOT_MESSAGE_QUEUE_SIZE;
use log::{error, info};
use prost::Message;
use raas_types::raas;
use raas_types::raas::bot::roll::Roll;
use raas_types::raas::cmd::request::Cmd;
use raas_types::raas::cmd::{Command, Request};
use raas_types::raas::register::{Register, RegisterResponse};
use raas_types::raas::resp::response::Resp;
use raas_types::raas::resp::Response;
use raas_types::raas::{recv_raas_msg, send_raas_msg, RaasMessage};
use std::collections::HashMap;
use std::ops::{Deref, DerefMut};
use std::time::{Duration, SystemTime, UNIX_EPOCH};
use tokio::io::{duplex, AsyncReadExt, AsyncWriteExt, DuplexStream};
use tokio::net::{TcpListener, TcpStream};
use tokio::select;
use tokio::time::timeout;

#[derive(Debug, Clone)]
pub enum RaaSCmd {
    Roll(u32),
    Img(Vec<u8>),
}

#[derive(Debug)]
struct ConnectorHandle {
    pub id: u32,
    pub stream: DuplexStream,
    pub tags: Vec<raas::register::BotTypes>,
}

#[derive(Debug)]
struct ConnectorManager {
    next_id: u32,
    connectors: HashMap<u32, ConnectorHandle>,
}

#[derive(Debug, Clone)]
pub struct RobotManager {
    addr: String,
    connector_manager: std::sync::Arc<tokio::sync::Mutex<ConnectorManager>>,
    command_stream: std::sync::Arc<tokio::sync::Mutex<DuplexStream>>,
}

impl RobotManager {
    pub fn new(addr: &str, stream: DuplexStream) -> Self {
        let connector_manager = ConnectorManager {
            next_id: 0,
            connectors: Default::default(),
        };

        Self {
            addr: addr.to_string(),
            connector_manager: std::sync::Arc::new(tokio::sync::Mutex::new(connector_manager)),
            command_stream: std::sync::Arc::new(tokio::sync::Mutex::new(stream)),
        }
    }

    async fn handle_register(&self, tcp_listener: &mut TcpListener) -> Result<(), std::io::Error> {
        info!("Listening for connection at {}", self.addr);
        let (mut socket, addr) = tcp_listener.accept().await?;
        info!("Got connection from {}", addr);

        let register_msg = recv_raas_msg(&mut socket).await?;

        let register = Register::decode(&*register_msg.msg).unwrap();

        let mut connector_manager = self.connector_manager.lock().await;

        let id = connector_manager.next_id;
        connector_manager.next_id += 1;

        let register_resp = RegisterResponse {
            name: register.name.to_string(),
            r#type: register.bot_type,
            id,
        };

        let mut msg = Vec::new();

        register_resp.encode(&mut msg).unwrap();

        send_raas_msg(&mut socket, msg).await?;

        let (connector_duplex_manager, connector_duplex_connector) =
            duplex(ROBOT_MESSAGE_QUEUE_SIZE);

        let mut connector = RobotConnector::new(
            id,
            vec![register.bot_type()],
            socket,
            connector_duplex_connector,
        );

        let connector_handler = ConnectorHandle {
            id,
            stream: connector_duplex_manager,
            tags: connector.bot_tags.clone(),
        };

        connector_manager.connectors.insert(id, connector_handler);

        tokio::spawn(async move { connector.worker().await });

        Ok(())
    }

    async fn recv_command(&self) -> Result<RaasMessage, Error> {
        let mut command_stream = self.command_stream.lock().await;

        let resp = recv_raas_msg(command_stream.deref_mut()).await?;

        Ok(resp)
    }

    async fn send_resp(&self, msg: RaasMessage) -> Result<(), Error> {
        let mut command_stream = self.command_stream.lock().await;

        command_stream.write_all(&*msg.into_bytes()).await?;

        Ok(())
    }

    pub async fn handle_command(&self) -> Result<(), Error> {
        info!("Waiting for command...");

        let msg = self.recv_command().await?;

        let request = Request::decode(&*msg.msg)?;

        let dest_type = match request.cmd.clone().unwrap() {
            Cmd::RollCmd(_) => raas::register::BotTypes::Roll,
        };

        let mut connector_manager = self.connector_manager.lock().await;
        let connector = connector_manager
            .connectors
            .iter_mut()
            .find(|(_, handler)| handler.tags.contains(&dest_type));

        if let Some((id, connector)) = connector {
            info!("Sending message to connector with id={}", id);
            let mut msg = Vec::new();
            request.encode(&mut msg)?;
            send_raas_msg(&mut connector.stream, msg).await?;

            let resp = recv_raas_msg(&mut connector.stream).await?;
            info!("Got response from id={}", id);
            self.send_resp(resp).await?;
        } else {
            error!("No connectors available to handle request");
        }

        Ok(())
    }

    pub async fn start_manager(self) -> Result<(), std::io::Error> {
        let handle_command_manager = self.clone();
        tokio::spawn(async move {
            loop {
                handle_command_manager.handle_command().await.unwrap()
            }
        });

        tokio::spawn(async move {
            let mut tcp_listener = TcpListener::bind(&self.addr).await.unwrap();
            loop {
                self.handle_register(&mut tcp_listener).await.unwrap()
            }
        });

        Ok(())
    }
}
